/*	$NetBSD: if_ieee80211.h,v 1.1 2000/01/23 23:50:14 chopps Exp $	*/

/*
 * Generic definitions for IEEE 802.11 frames, from IEEE802.11 Draft 61 1997
 *
 * IEEE 802.11 Address Field Assignments
 *
 *              FROMDS TODS  i_addr1    i_addr2    i_addr3    i_addr4
 * Sta to Sta   0      0     Dest STA   Src  STA   BSSID      N/A
 * Sta to AP    0      1     BSSID      Src  STA   Dest AP    N/A
 * AP to Sta    1      0     Dest STA   BSSID      Src  AP    N/A
 * AP to AP     1      1     Dest  AP   Src   AP   Dest STA   Src STA
 *
 * The BSSID is a MAC layer address so i_addr1 becomes the address of the
 * immediate recipient of the packets, and i_addr2 is the address of the
 * packet sender.
 *
 */
struct ieee80211_header {
	u_int8_t	i_fc[2];
	u_int8_t	i_dur[2];
	u_int8_t	i_addr1[ETHER_ADDR_LEN];
	u_int8_t	i_addr2[ETHER_ADDR_LEN];
	u_int8_t	i_addr3[ETHER_ADDR_LEN];
	u_int8_t	i_seq[2];
	/* possibly followed by addr4[ETHER_ADDR_LEN]; */
};

#define	IEEE80211_FC0_VERSION_MASK		0x03
#define	IEEE80211_FC0_VERSION_0			0x00

#define	IEEE80211_FC0_TYPE_MASK			0x0c
#define	IEEE80211_FC0_SUBTYPE_MASK		0xf0
#define	IEEE80211_FC0_TYPE_MGT			0x00
#define	IEEE80211_FC0_TYPE_CTL			0x04
#define	IEEE80211_FC0_TYPE_DATA			0x08

#define	IEEE80211_FC0_SUBTYPE_MGT_ASSOC_REQ	0x00
#define	IEEE80211_FC0_SUBTYPE_MGT_ASSOC_RESP	0x10
#define	IEEE80211_FC0_SUBTYPE_MGT_REASSOC_REQ	0x20
#define	IEEE80211_FC0_SUBTYPE_MGT_REASSOC_RESP	0x30
#define	IEEE80211_FC0_SUBTYPE_MGT_PROBE_REQ	0x40
#define	IEEE80211_FC0_SUBTYPE_MGT_PROBE_RESP	0x50
#define	IEEE80211_FC0_SUBTYPE_MGT_RESERVED0	0x60
#define	IEEE80211_FC0_SUBTYPE_MGT_RESERVED1	0x70

#define	IEEE80211_FC0_SUBTYPE_MGT_BEACON	0x80
#define	IEEE80211_FC0_SUBTYPE_MGT_AUTH_RESP	0x90
#define	IEEE80211_FC0_SUBTYPE_MGT_DISASSOC	0xa0
#define	IEEE80211_FC0_SUBTYPE_MGT_AUTH		0xb0
#define	IEEE80211_FC0_SUBTYPE_MGT_DEAUTH	0xc0
#define	IEEE80211_FC0_SUBTYPE_MGT_RESERVED2	0xd0
#define	IEEE80211_FC0_SUBTYPE_MGT_RESERVED3	0xe0
#define	IEEE80211_FC0_SUBTYPE_MGT_RESERVED4	0xf0

#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED0	0x00
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED1	0x10
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED2	0x20
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED3	0x30
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED4	0x40
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED5	0x50
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED6	0x60
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED7	0x70

#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED8	0x80
#define	IEEE80211_FC0_SUBTYPE_CTL_RESERVED9	0x90
#define	IEEE80211_FC0_SUBTYPE_CTL_PS_POLL	0xa0
#define	IEEE80211_FC0_SUBTYPE_CTL_RTS		0xb0
#define	IEEE80211_FC0_SUBTYPE_CTL_CTS		0xc0
#define	IEEE80211_FC0_SUBTYPE_CTL_ACK		0xd0
#define	IEEE80211_FC0_SUBTYPE_CTL_CFEND		0xe0
#define	IEEE80211_FC0_SUBTYPE_CTL_CFEND_CFACK	0xf0

#define	IEEE80211_FC0_SUBTYPE_DATA_DATA		0x00
#define	IEEE80211_FC0_SUBTYPE_DATA_DATA_CFACK	0x10
#define	IEEE80211_FC0_SUBTYPE_DATA_DATA_CFPOLL	0x20
#define	IEEE80211_FC0_SUBTYPE_DATA_DATA_CFACPL	0x30
#define	IEEE80211_FC0_SUBTYPE_DATA_NULL		0x40
#define	IEEE80211_FC0_SUBTYPE_DATA_CFACK	0x50
#define	IEEE80211_FC0_SUBTYPE_DATA_CFPOLL	0x60
#define	IEEE80211_FC0_SUBTYPE_DATA_CFACK_CFACK	0x70

#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED0	0x80
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED1	0x90
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED2	0xa0
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED3	0xb0
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED4	0xc0
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED5	0xd0
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED6	0xe0
#define	IEEE80211_FC0_SUBTYPE_DATA_RESERVED7	0xf0

#define IEEE80211_FC1_TODS			0x01
#define IEEE80211_FC1_FROMDS			0x02
#define IEEE80211_FC1_MORE_FRAG			0x04
#define IEEE80211_FC1_RETRY			0x08
#define IEEE80211_FC1_PWR_MGT			0x10
#define IEEE80211_FC1_MORE_DATAi		0x20
#define IEEE80211_FC1_WEP			0x40
#define IEEE80211_FC1_ORDER			0x80

#define	IEEE80211_FC1_DS_MASK \
			(IEEE80211_FC1_TODS | IEEE80211_FC1_FROMDS)
#define	IEEE80211_FC1_STA_TO_STA		0x00
#define	IEEE80211_FC1_STA_TO_AP			IEEE80211_FC1_TODS
#define	IEEE80211_FC1_AP_TO_STA			IEEE80211_FC1_FROMDS
#define	IEEE80211_FC1_AP_TO_AP \
			(IEEE80211_FC1_TODS | IEEE80211_FC1_FROMDS)

#define	IEEE80211_NWID_LEN	32

/* nwid is u_int8_t array of IEEE80211_NWID_LEN pointed at by ifr.ifr_data */
#define	SIOCS80211NWID		_IOWR('i', 230, struct ifreq)
#define	SIOCG80211NWID		_IOWR('i', 231, struct ifreq)
